import os
import re
import ast as A
from enum import Enum
from functools import lru_cache
from typing import List, Set, Union
from hashlib import sha256 as _hash

def sha256(x: str):
  """generic function to calculate SHA256 hash of a string"""
  return _hash(x.encode()).hexdigest()

class IndexTypes(Enum):
  """
  This is our list of index types, these can be used to search through things in the AST. Here's some description on
  what each of these mean. There are a few special ones for convinience:
  - FILE: which means that the current location is the file itself, also can be considered as ast.Module
  - DOCSTRING: which is a special type of expression that is a docstring, is defined to use a different parsing logic
  """
  FUNCTION = A.FunctionDef
  CLASS = A.ClassDef
  VARIABLE = A.Name
  IMPORT = A.Import
  IMPORT_FROM = A.ImportFrom
  ASSIGN = A.Assign
  EXPRESSION = A.Expr
  IF = A.If
  FILE = "file"
  DOCSTRING = "__doc__"

  def all():
    """returns all the valid index types"""
    return [
      IndexTypes.FUNCTION,
      IndexTypes.CLASS,
      IndexTypes.VARIABLE,
      IndexTypes.IMPORT,
      IndexTypes.IMPORT_FROM,
      IndexTypes.ASSIGN,
      IndexTypes.EXPRESSION,
      IndexTypes.IF,
      IndexTypes.FILE
    ]


def get_code_section(node: A.AST, code_lines: List[str]):
  """generic function to return the code section for a node"""
  if type(code_lines) == str:
    code_lines = code_lines.splitlines()
  else:
    assert type(code_lines) == list and type(code_lines[0]) == str
  sl, so, el, eo = node.lineno, node.col_offset, node.end_lineno, node.end_col_offset
  code = ""
  if sl == el:
    code = code_lines[sl-1][so:eo]
  else:
    for i in range(sl - 1, el, 1):
      if i == sl - 1:
        code += code_lines[i][so:]
      elif i == el - 1:
        code += "\n" + code_lines[i][:eo]
      else:
        code += "\n" + code_lines[i]
  return code



class Astea:
  """### Engineer's Note
  
  Astea is designed from ground up to suit needs of code traversal and search. It's not a full AST engine, but it's
  good enough for our needs. In the first version, we wrote `nbox.framework.on_functions` which was written to generate
  DAGs for NBX-JobsFlow, it was overly complicated and not well engineered for great UX. So we ended up writing
  something brand new which reflects more like how humans work traverse code. This is very important because down the
  road we want to do really insane and powerful things with code.
  """
  def __init__(
    self,
    fname: str = "",
    code: str = "",
    *,
    name: str = "",
    type: IndexTypes = None,
    node: A.AST = None,
    code_lines: List[str] = None,
    order_index: int = -1,
  ) -> None:
    """This is an AST node, that can be called upon itself to traverse in a natural human way. Only `fname` is an important
    argument, others will be populated automatically.

    Examples:
      # Creating a simple tea:
      >>> from nbox.nbxlib import astea as A
      >>> tea = A.Astea(fname = A.__file__)

    Args:
      fname (str): the file name to parse. If this is not provided, then all other values must be provided.
      code (str): the code text can be provided instead of the.
      name (str, optional): The name of the node. Defaults to "".
      type (IndexTypes, optional): The type of the node. Defaults to None.
      node (A.AST, optional): The AST node. Defaults to None.
      code_lines (List[str], optional): The code lines. Defaults to None.
      order_index (int, optional): The order index of the node. Defaults to -1.
    """

    if fname:
      assert os.path.exists(fname), f"file/folder {fname} does not exist"
      if os.path.isdir(fname):
        fname = os.path.join(fname, "__init__.py")
      type = IndexTypes.FILE
      with open(fname, 'r') as f:
        code = f.read()
        node: A.Module = A.parse(code)
      code_lines = code.splitlines()
      name = fname
    elif code:
      type = IndexTypes.FILE
      node: A.Module = A.parse(code)
      code_lines = code.splitlines()
      name = "0"
    elif name:
      pass
    else:
      raise ValueError("Must provide either fname/code or name")

    self.name = name
    self.type = type
    self.node = node
    self.code_lines = code_lines
    self.order_index = order_index

    if hasattr(node, "lineno"):
      self._code = get_code_section(node, code_lines)
    else:
      self._code = "\n".join(self.code_lines)
    self._sha256 = sha256(self._code)

  def __repr__(self) -> str:
    return f"{self.type.name} {self.name} {self._sha256[:6]}"

  def __hash__(self) -> int:
    # since we know that eventually the sha256 of the code is the only unique identifier for nay piece of text
    # we can return the first 16 bytes of the sha256 as the hash, this is also useful since same object in two
    # different python processes will have different hashes, but the sha256 will always be the same.
    return int(self._sha256[:16], 16)

  @lru_cache()
  def docstring(self) -> str:
    """Get the docstring of the current node, is valid only for functions and classes. This will automatically
    perform relevant dedentation and return the docstring as a single string."""
    docstring = ""
    if type(self.node) in [A.FunctionDef, A.ClassDef, A.Module]:
      # check if this has a documentation string
      if self.node.body and isinstance(self.node.body[0], A.Expr) and isinstance(self.node.body[0].value, A.Str):
        docstring = self.node.body[0].value.s
    else:
      raise ValueError(f"Can't get docstring for {self.name} ({self.type})")
    docstring = docstring.strip()

    # in some cases we need to remove the indentation from the docstring
    doc_lines = docstring.splitlines()
    if len(doc_lines) > 1:
      _l = doc_lines[1]
      _i = 1
      while not _l and _i < len(doc_lines):
        _i += 1
        _l = doc_lines[_i]
      indent = len(_l) - len(_l.lstrip())
      docstring = doc_lines[0] + "\n" + "\n".join([line[indent:] for line in doc_lines[1:]])
    return docstring

  @property
  @lru_cache(1)
  def index(self) -> List['Astea']:
    """Returns a list of all the AST nodes (same `Astea`) in the current node. This is a recursive function and user/bot
    is expected to go over this to traverse."""
    items = set()
    if hasattr(self.node, 'body'):
      for i, n in enumerate(self.node.body):
        if type(n) == A.FunctionDef:
          tea = Astea(name=n.name, type=IndexTypes.FUNCTION, node=n, code_lines=self.code_lines, order_index=i)
          items.add(tea)
        elif type(n) == A.ClassDef:
          tea = Astea(name=n.name, type=IndexTypes.CLASS, node=n, code_lines=self.code_lines, order_index=i)
          items.add(tea)
        elif type(n) == A.Import:
          for j, nn in enumerate(n.names):
            tea = Astea(name=nn.name, type=IndexTypes.IMPORT, node=n, code_lines=self.code_lines, order_index=i+j)
            items.add(tea)
        elif type(n) == A.ImportFrom:
          for j, nn in enumerate(n.names):
            tea = Astea(name=nn.name, type=IndexTypes.IMPORT_FROM, node=n, code_lines=self.code_lines, order_index=i+j)
            items.add(tea)
        elif type(n) == A.Expr:
          tea = Astea(name=n.value, type=IndexTypes.EXPRESSION, node=n, code_lines=self.code_lines, order_index=i)
          items.add(tea)
        
        # if this node is a class and node 'n' is a staticmethod then assign is considered
        elif type(n) == A.Assign:
          target = n.targets[0]
          if type(target) == A.Tuple:
            # iterate over the tuple and add each item as VARIABLE
            for j, nn in enumerate(target.elts):
              tea = Astea(name=nn.id, type=IndexTypes.VARIABLE, node=n, code_lines=self.code_lines, order_index=i+j)
              items.add(tea)
            continue
          elif type(target) == A.Name:
            tea = Astea(name=target.id, type=IndexTypes.VARIABLE, node=n, code_lines=self.code_lines, order_index=i)
            items.add(tea)
          else:
            # raise ValueError(f"Unknown type {type(target)} for target {target}")
            pass
          if (type(self.node) == A.ClassDef or type(self.node) == A.Module):
            try:
              # this .id is very painful
              tea = Astea(name=n.targets[0].id, type=IndexTypes.ASSIGN, node=n, code_lines=self.code_lines, order_index=i)
              items.add(tea)
            except:
              # though we eventually want to get away with this try/except, we will keep it just in case
              pass
        else:
          continue
    items = sorted(list(items), key=lambda x: x.order_index)
    return items

  def find(self, x: str = "", types: Union[IndexTypes, List[IndexTypes]] = None) -> List['Astea']:
    """Find all the instances of x in the current node, and return a list of IndexItems.
    
    - if there is a '.' (dot) in `x`, `find` will try to perform a recursive search, you can escape the dot with a backslash
    - This currently does not work correctly when the same name is used in different types
    - if both `x` and `types` is not provided, then this behaves just like `.index`

    Examples:

      # Create a tea:
      >>> from nbox.nbxlib import astea as A
      >>> tea = A.Astea(fname = A.__file__)

      # Find things inside the tea:
      >>> cls_obj = tea.find("Astea")[0]      # search by name 
      >>> fn_obj = cls_obj.find('find')       # search inside any Astea
      >>> fn_obj = tea.find("Astea.find")[0]  # search inside things by using . (dot)

    Args:
      x (str): If provided, will find the name of the item to search for
      types (Union[IndexTypes, List[IndexTypes]], optional): The type of items to search for. Defaults to None.

    Returns:
      List['Astea']: A list of IndexItems that match the search
    """
    if types is None:
      types = IndexTypes.all()
    elif isinstance(types, IndexTypes):
      types = [types]

    # get the sections of the search string
    sections = []
    if x:
      for s in re.split(r'(?<!\\)\.', x):
        sections.append(s.replace(r'\.', '.'))

    res = []
    node = self
    if sections:
      for i, s in enumerate(sections):
        if i == len(sections) - 1:
          # last section, we need to find the actual item
          for item in node.index:
            if item.name == s and item.type in types:
              res.append(item)
        else:
          # we need to find the next node
          for item in node.index:
            if item.name == s and item.type in types:
              node = item
              break
          else:
            # we didn't find the node, so we can't continue
            break
    else:
      # no sections, so we need to return all the items in the current node
      for item in node.index:
        if item.type in types:
          res.append(item)
    return res



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import common.common_pb2
import google.protobuf.descriptor
import google.protobuf.field_mask_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Serving(google.protobuf.message.Message):
    """
    A Serving (Group) is a collection of Model(s).
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    MODELS_FIELD_NUMBER: builtins.int
    DEPLOYMENT_ENDPOINT_FIELD_NUMBER: builtins.int
    MODEL_ENDPOINTS_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    CPU_SCALE_CONFIG_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    id: builtins.str
    """ID for the Serving Group"""
    name: builtins.str
    """name of the Serving Group"""
    @property
    def models(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Model]:
        """list of models"""
    @property
    def deployment_endpoint(self) -> global___Endpoint:
        """Endpoint for the Serving Group"""
    @property
    def model_endpoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Endpoint]:
        """Endpoints for the models. Copy of models[*].endpoint."""
    @property
    def resource(self) -> common.common_pb2.Resource: ...
    @property
    def cpu_scale_config(self) -> global___ScaleConfig: ...
    description: builtins.str
    """description of the serving group"""
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """metadata for serving group"""
    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        name: builtins.str = ...,
        models: collections.abc.Iterable[global___Model] | None = ...,
        deployment_endpoint: global___Endpoint | None = ...,
        model_endpoints: collections.abc.Iterable[global___Endpoint] | None = ...,
        resource: common.common_pb2.Resource | None = ...,
        cpu_scale_config: global___ScaleConfig | None = ...,
        description: builtins.str = ...,
        metadata: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        created_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["cpu_scale_config", b"cpu_scale_config", "created_at", b"created_at", "deployment_endpoint", b"deployment_endpoint", "resource", b"resource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["cpu_scale_config", b"cpu_scale_config", "created_at", b"created_at", "deployment_endpoint", b"deployment_endpoint", "description", b"description", "id", b"id", "metadata", b"metadata", "model_endpoints", b"model_endpoints", "models", b"models", "name", b"name", "resource", b"resource"]) -> None: ...

global___Serving = Serving

@typing_extensions.final
class Endpoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    ACCESS_KEY_FIELD_NUMBER: builtins.int
    HOST_FIELD_NUMBER: builtins.int
    id: builtins.str
    """Model or Serving (Group) ID"""
    url: builtins.str
    """URL to access the model or serving group"""
    access_key: builtins.str
    """access_key is used to authenticate requests to the endpoint.
    It is used in the following ways:
    curl -H "NBX-KEY:  $access_key" https://api.nimblebox.ai/ |  $url
    """
    host: builtins.str
    """Hostname of the endpoint defined in the apisix route"""
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        url: builtins.str = ...,
        access_key: builtins.str = ...,
        host: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["access_key", b"access_key", "host", b"host", "id", b"id", "url", b"url"]) -> None: ...

global___Endpoint = Endpoint

@typing_extensions.final
class ServingStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Status:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _StatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ServingStatus._Status.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        STATUS_UNSPECIFIED: ServingStatus._Status.ValueType  # 0
        """Unknown status"""
        STATUS_PENDING: ServingStatus._Status.ValueType  # 1
        """Deployment is pending"""
        STATUS_SUCCESS: ServingStatus._Status.ValueType  # 2
        """Deployment is successful"""
        STATUS_FAILED: ServingStatus._Status.ValueType  # 3
        """Deployment is failed"""

    class Status(_Status, metaclass=_StatusEnumTypeWrapper): ...
    STATUS_UNSPECIFIED: ServingStatus.Status.ValueType  # 0
    """Unknown status"""
    STATUS_PENDING: ServingStatus.Status.ValueType  # 1
    """Deployment is pending"""
    STATUS_SUCCESS: ServingStatus.Status.ValueType  # 2
    """Deployment is successful"""
    STATUS_FAILED: ServingStatus.Status.ValueType  # 3
    """Deployment is failed"""

    UPLOAD_STATUS_FIELD_NUMBER: builtins.int
    DEPLOY_STATUS_FIELD_NUMBER: builtins.int
    upload_status: global___ServingStatus.Status.ValueType
    deploy_status: global___ServingStatus.Status.ValueType
    def __init__(
        self,
        *,
        upload_status: global___ServingStatus.Status.ValueType = ...,
        deploy_status: global___ServingStatus.Status.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["deploy_status", b"deploy_status", "upload_status", b"upload_status"]) -> None: ...

global___ServingStatus = ServingStatus

@typing_extensions.final
class ScaleConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MIN_REPLICAS_FIELD_NUMBER: builtins.int
    MAX_REPLICAS_FIELD_NUMBER: builtins.int
    THRESHOLD_FIELD_NUMBER: builtins.int
    min_replicas: builtins.int
    max_replicas: builtins.int
    threshold: builtins.float
    def __init__(
        self,
        *,
        min_replicas: builtins.int = ...,
        max_replicas: builtins.int = ...,
        threshold: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["max_replicas", b"max_replicas", "min_replicas", b"min_replicas", "threshold", b"threshold"]) -> None: ...

global___ScaleConfig = ScaleConfig

@typing_extensions.final
class ServingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    @property
    def serving(self) -> global___Serving: ...
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        serving: global___Serving | None = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "serving", b"serving"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "serving", b"serving"]) -> None: ...

global___ServingRequest = ServingRequest

@typing_extensions.final
class UpdateServingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_FIELD_NUMBER: builtins.int
    UPDATE_MASK_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    @property
    def serving(self) -> global___Serving: ...
    @property
    def update_mask(self) -> google.protobuf.field_mask_pb2.FieldMask: ...
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        serving: global___Serving | None = ...,
        update_mask: google.protobuf.field_mask_pb2.FieldMask | None = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "serving", b"serving", "update_mask", b"update_mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "serving", b"serving", "update_mask", b"update_mask"]) -> None: ...

global___UpdateServingRequest = UpdateServingRequest

@typing_extensions.final
class ServingListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    limit: builtins.int
    offset: builtins.int
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        limit: builtins.int = ...,
        offset: builtins.int = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "limit", b"limit", "offset", b"offset"]) -> None: ...

global___ServingListRequest = ServingListRequest

@typing_extensions.final
class ServingListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERVINGS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_FIELD_NUMBER: builtins.int
    @property
    def servings(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Serving]: ...
    next_page: builtins.int
    def __init__(
        self,
        *,
        servings: collections.abc.Iterable[global___Serving] | None = ...,
        next_page: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["next_page", b"next_page", "servings", b"servings"]) -> None: ...

global___ServingListResponse = ServingListResponse

@typing_extensions.final
class Model(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _ServingType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ServingTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Model._ServingType.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        SERVING_TYPE_UNSPECIFIED: Model._ServingType.ValueType  # 0
        """Unknown deployment type"""
        SERVING_TYPE_NBOX_OP: Model._ServingType.ValueType  # 1
        """Serving via nbox Operators"""

    class ServingType(_ServingType, metaclass=_ServingTypeEnumTypeWrapper):
        """Deployment type"""

    SERVING_TYPE_UNSPECIFIED: Model.ServingType.ValueType  # 0
    """Unknown deployment type"""
    SERVING_TYPE_NBOX_OP: Model.ServingType.ValueType  # 1
    """Serving via nbox Operators"""

    class _PinStatus:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _PinStatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Model._PinStatus.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        PIN_STATUS_UNSPECIFIED: Model._PinStatus.ValueType  # 0
        """Unknown pin status"""
        PIN_STATUS_PINNED: Model._PinStatus.ValueType  # 1
        """Model is pinned"""
        PIN_STATUS_UNPINNED: Model._PinStatus.ValueType  # 2
        """Model is unpinned"""

    class PinStatus(_PinStatus, metaclass=_PinStatusEnumTypeWrapper): ...
    PIN_STATUS_UNSPECIFIED: Model.PinStatus.ValueType  # 0
    """Unknown pin status"""
    PIN_STATUS_PINNED: Model.PinStatus.ValueType  # 1
    """Model is pinned"""
    PIN_STATUS_UNPINNED: Model.PinStatus.ValueType  # 2
    """Model is unpinned"""

    @typing_extensions.final
    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class FeatureGatesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    SERVING_GROUP_ID_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    CODE_FIELD_NUMBER: builtins.int
    ENDPOINT_FIELD_NUMBER: builtins.int
    PIN_STATUS_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    REPLICAS_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    K8SINFO_FIELD_NUMBER: builtins.int
    FEATURE_GATES_FIELD_NUMBER: builtins.int
    id: builtins.str
    """ID for the model"""
    name: builtins.str
    """name of the model"""
    serving_group_id: builtins.str
    """ID of the serving group the model belongs to"""
    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time created"""
    type: global___Model.ServingType.ValueType
    @property
    def resource(self) -> common.common_pb2.Resource:
        """model resource"""
    @property
    def code(self) -> common.common_pb2.Code: ...
    @property
    def endpoint(self) -> global___Endpoint: ...
    pin_status: global___Model.PinStatus.ValueType
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """metadata for model"""
    replicas: builtins.int
    """number of replicas for model"""
    @property
    def status(self) -> global___ServingStatus: ...
    @property
    def k8sInfo(self) -> common.common_pb2.K8sInfo: ...
    @property
    def feature_gates(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        name: builtins.str = ...,
        serving_group_id: builtins.str = ...,
        created_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        type: global___Model.ServingType.ValueType = ...,
        resource: common.common_pb2.Resource | None = ...,
        code: common.common_pb2.Code | None = ...,
        endpoint: global___Endpoint | None = ...,
        pin_status: global___Model.PinStatus.ValueType = ...,
        metadata: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        replicas: builtins.int | None = ...,
        status: global___ServingStatus | None = ...,
        k8sInfo: common.common_pb2.K8sInfo | None = ...,
        feature_gates: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_replicas", b"_replicas", "code", b"code", "created_at", b"created_at", "endpoint", b"endpoint", "k8sInfo", b"k8sInfo", "replicas", b"replicas", "resource", b"resource", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_replicas", b"_replicas", "code", b"code", "created_at", b"created_at", "endpoint", b"endpoint", "feature_gates", b"feature_gates", "id", b"id", "k8sInfo", b"k8sInfo", "metadata", b"metadata", "name", b"name", "pin_status", b"pin_status", "replicas", b"replicas", "resource", b"resource", "serving_group_id", b"serving_group_id", "status", b"status", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_replicas", b"_replicas"]) -> typing_extensions.Literal["replicas"] | None: ...

global___Model = Model

@typing_extensions.final
class ModelRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    @property
    def model(self) -> global___Model: ...
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        model: global___Model | None = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "model", b"model"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "model", b"model"]) -> None: ...

global___ModelRequest = ModelRequest

@typing_extensions.final
class UpdateModelRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    UPDATE_MASK_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    @property
    def model(self) -> global___Model: ...
    @property
    def update_mask(self) -> google.protobuf.field_mask_pb2.FieldMask: ...
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        model: global___Model | None = ...,
        update_mask: google.protobuf.field_mask_pb2.FieldMask | None = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "model", b"model", "update_mask", b"update_mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "model", b"model", "update_mask", b"update_mask"]) -> None: ...

global___UpdateModelRequest = UpdateModelRequest

@typing_extensions.final
class ModelListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_ID_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    INTEGRATION_DATA_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common.common_pb2.NBXAuthInfo: ...
    serving_id: builtins.str
    limit: builtins.int
    offset: builtins.int
    @property
    def integration_data(self) -> common.common_pb2.IntegrationData: ...
    def __init__(
        self,
        *,
        auth_info: common.common_pb2.NBXAuthInfo | None = ...,
        serving_id: builtins.str = ...,
        limit: builtins.int = ...,
        offset: builtins.int = ...,
        integration_data: common.common_pb2.IntegrationData | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "integration_data", b"integration_data", "limit", b"limit", "offset", b"offset", "serving_id", b"serving_id"]) -> None: ...

global___ModelListRequest = ModelListRequest

@typing_extensions.final
class ModelListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODELS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_FIELD_NUMBER: builtins.int
    @property
    def models(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Model]: ...
    next_page: builtins.int
    def __init__(
        self,
        *,
        models: collections.abc.Iterable[global___Model] | None = ...,
        next_page: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["models", b"models", "next_page", b"next_page"]) -> None: ...

global___ModelListResponse = ModelListResponse



"""
`Operators` is how you write NBX-Jobs. If you are familiar with pytorch, then usage is
exactly same, for others here's a quick recap:

```python
class MyOperator(Operator):
  def __init__(self, a: int, b: str):
    super().__init__()
    self.a: int = a
    self.b: Operator = MyOtherOperator(b) # nested calling
  
  def forward(self, x: int) -> int:
    y = self.a + x
    y = self.b(y) + x # nested calling
    return y

job = MyOperator(1, "hello") # define once
res = job(2)                 # use like python, screw DAGs
```

We always wanted to ensure that there is least developer resistance in the way of using `Operator`
so there is a convinient `operator` decorator that can wrap any function or class and extend all
the powerful methods available in the `Operator` object, like `.deploy()`. By default every
wrapped function is run as a Job.

```python
@operator()
def foo(i: float = 4):
  return i * i

# to deploy the operator 
if __name__ == "__main__":
  # pass deployment_type = "serving" to make an API
  foo_remote = foo.deploy('workspace-id')
  assert foo_remote() == foo()
  assert foo_remote(10) == foo(10)
```

And you can make simple stateful object like classes using `@operator` decorator by making it
an API endpoint.

```python
@operator()
class Bar:
  def __init__(self, x: int = 1):
    self.x = x

  def inc(self):
    self.x += 1

  def getvalue(self):
    return self.x

  def __getattr__(self, k: str):
    # simple echo to demonstrate that underlying python object methods can
    # also be accessed over the internet
    return str

if __name__ == "__main__":
  bar_remote = Bar.deploy('workspace-id')
  
  # increment the numbers
  bar.inc(); bar_remote.inc()

  # directly access the values, no need for futures
  assert bar.x == bar_remote.x

  print(bar.jj_guverner, bar_remote.jj_guverner)
```

If you want to use the APIs for deployed jobs and servings (nbox.Jobs)[nbox.jobs.html] is a better documentation.

## Engineering

Fundamentally operators act as a wrapper on user code, sometime abstracting away functions
by breaking them into `__init__`s and `forward`s. But this is a simpler way to wrap
user function than letting users wrap their own function. It is easy to get false positives,
and so we explicitly expand things in two. These operators are like `torch.nn.Modules`
spiritually as well because modules manage the underlying weights and operators manage the
underlying user logic.

Operators are combination of several subsystems that are all added in the same class, though
certainly if we come up with that high abstraction we will refactor this:

- tree: All operators are really treated like a tree meaning that the execution is nested and the order of execution is
  determined by the order of the operators in the tree. DAGs are fundamentally just trees with some nodes spun together,
  to execute only once.
- deploy, ...: All the services in NBX-Jobs.
- get_nbx_flow: which is the static code analysis system to understand true user intent and if possible (and permission of
  the user) optimise the logic.

## Tips

`Operators` are built to be the abstract equivalent of any computation so code can be easily
run in distributed fashion.

- Use Operator directly as a function as much as possible, it's the simplest way to use it.
- `@operator` decorator on your function and it will be run as a job by default, you want that.
- `@operator` decorator on your class and it will be run as a serving by default, you want that.

## Documentation

"""
# Some parts of the code are based on the pytorch nn.Module class
# pytorch license: https://github.com/pytorch/pytorch/blob/master/LICENSE
# modifications: research@nimblebox.ai 2022

from functools import partial
import json
import os
import re
import inspect
import requests
from tqdm import tqdm
from functools import partial
from time import sleep, monotonic
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, Iterable, List, Tuple, Union

import nbox.utils as U
from nbox.auth import secret, ConfigString
from nbox.utils import logger, SimplerTimes
from nbox.nbxlib import operator_spec as ospec
from nbox.nbxlib.tracer import Tracer
from nbox.version import __version__
from nbox.sub_utils.latency import log_latency
from nbox.framework.on_functions import get_nbx_flow
from nbox.framework import AirflowMixin, PrefectMixin
from nbox.hyperloop.jobs.job_pb2 import Job as JobProto
from nbox.hyperloop.common.common_pb2 import Resource
from nbox.hyperloop.jobs.dag_pb2 import DAG, Flowchart, Node, RunStatus

from nbox.jobs import Schedule, Job, Serve, _get_job_data
from nbox.messages import write_binary_to_file
from nbox.relics import Relics
from nbox.init import nbox_ws_v1
from nbox.subway import SpecSubway


# alias
DEFAULT_RESOURCE = ospec.DEFAULT_RESOURCE
FN_IGNORE = ospec.FN_IGNORE

class Operator():
  node = Node()
  source_edges: List[Node] = None
  _tracer: Tracer = None
  
  # there are some thing that are used to hide the complexity of the entire operation
  # from the user
  _op_type = ospec.OperatorType.UNSET
  _op_spec: Union[ospec._UnsetSpec,ospec._JobSpec,ospec._ServingSpec,ospec._WrapClsSpec,ospec._WrapFnSpec] = ospec._UnsetSpec()

  # this is a map from operator to resource, by deafult the values will be None,
  # unless explicitly modified by `chief.machine.Machine`
  _current_resource = None
  _op_to_resource_map = {}

  def __init__(self) -> None:
    """Create an operator, which abstracts your code into sharable, bulding blocks which
    can then deployed on either NBX-Jobs or NBX-Deploy.
    
    Examples:

      class MyOperator(Operator):
        def __init__(self, ...):
          ... # initialisation of job happens here

          # use prebuilt operators to define the entire process
          from nbox.lib.shell import Shell
          self.download_binary = Shell("wget https://nbox.ai/{hash}")

          # keep a library of organisation wide operators
          from .nbx_internal.operators import TrainGPT
          self.train_model: Operator = TrainGPT(...)

        def forward(self, ...):
          # pass any inputs you want at runtime
          ... # execute code in any arbitrary order, free from DAGs

          self.download_binary(hash="my_binary") # pass relevant data at runtime
          self.train_model() # run any operation and get full visibility on platform

      # to convert operator is 
      job: Operator = MyOperator(...)

      # deploy this as a batch process or API endpoint or API endpoint
      job.deploy() # WIP # WIP
    """
    # do not add anything in here, else it will call self.__getattr__ and cause a RecursionError
    self._operators: Dict[str, 'Operator'] = OrderedDict() # {name: operator}

  def __remote_init__(self):
    """User can overwrite this function, this will be called only when running on remote.
    This helps in with things like creating the models can caching them in self, instead
    of `lru_cache` in forward."""
    pass

  def remote_init(self):
    """Triggers `__remote_init__` across the entire tree."""
    self.__remote_init__()
    for _, op in self._operators.items():
      op.remote_init()

  def __repr__(self):
    # from torch.nn.Module
    def _addindent(s_, numSpaces):
      s = s_.split('\n')
      # don't do anything for single-line stuff
      if len(s) == 1:
        return s_
      first = s.pop(0)
      s = [(numSpaces * ' ') + line for line in s]
      s = '\n'.join(s)
      s = first + '\n' + s
      return s

    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    extra_repr = ""
    # empty string will be split into list ['']
    if extra_repr:
      extra_lines = extra_repr.split('\n')
    child_lines = []
    for key, module in self._operators.items():
      mod_str = repr(module)
      mod_str = _addindent(mod_str, 2)
      child_lines.append('(' + key + '): ' + mod_str)
    lines = extra_lines + child_lines

    main_str = self.__qualname__ + '('
    if lines:
      # simple one-liner info, which most builtin Modules will use
      if len(extra_lines) == 1 and not child_lines:
        main_str += extra_lines[0]
      else:
        main_str += '\n  ' + '\n  '.join(lines) + '\n'

    main_str += ')'
    return main_str

  # mixin/

  # to_airflow_operator = AirflowMixin.to_airflow_operator
  # to_airflow_dag = AirflowMixin.to_airflow_dag
  # from_airflow_operator = classmethod(AirflowMixin.from_airflow_operator)
  # from_airflow_dag = classmethod(AirflowMixin.from_airflow_dag)
  # to_prefect_task = PrefectMixin.to_prefect_task
  # to_prefect_flow = PrefectMixin.to_prefect_flow
  # from_prefect_task = classmethod(PrefectMixin.from_prefect_task)
  # from_prefect_flow = classmethod(PrefectMixin.from_prefect_flow)

  # Operators are very versatile and can be created in many different ways, heres a list of those:
  # 1. from an existing job, where calling it triggers a new run and args are passed through relics
  # 2. from an existing serving, where calling it equates to an HTTP API call
  # 3. from a function, where we decorate an existing function and convert it to an operator

  @classmethod
  def from_job(cls, job_name: str = "", job_id: str = "", workspace_id: str = ""):
    """latch an existing job so that it can be called as an operator. This is designed to be used as
    a part of "Compute Fabric" mode.

    Args:
      job_name (str, optional): Name of the job. Defaults to "".
      job_id (str, optional): ID of the job. Defaults to "".
    """
    # implement this when we have the client-server that allows us to get the metadata for the job
    if (not job_name and not job_id) or (job_name and job_id):
      raise ValueError("Either job_name or job_id must be specified")
    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
    if not workspace_id:
      raise DeprecationWarning("workspace_id cannot be none. Reauth yourself with `nbx login`")
    job_id, job_name = _get_job_data(
      name = job_name,
      id = job_id,
      remove_archived = True, 
      workspace_id = workspace_id
    )
    if job_id is None:
      raise ValueError(f"No serving found with name {job_name}")
    job = Job(job_id = job_id, workspace_id = workspace_id)
    logger.debug(f"Latching to job '{job_name}' ({job_id})")

    def forward(*args, _wait: bool = True, **kwargs):
      """This is the forward method for a NBX-Job. All the parameters will be passed through Relics."""
      logger.debug(f"Running job '{job_name}' ({job_id})")
      relic = Relics("cache", workspace_id, create = True)

      # determining the put location is very tricky because there is no way to create a sync between the
      # key put here and what the run will pull. This can lead to many weird race conditions. So for now
      # I am going to rely on the fact that we cannot have two parallel runs with same tag. Thus at any given
      # moment there can be only one file at /{job_id}/args_kwargs_{tag}.pkl
      tag = U.get_random_name(True).split("-")[0]
      relic.put_object(f"{job_id}/args_kwargs_{tag}", (args, kwargs))

      # and then we will trigger the job with that specific tag and wait for the run to complete
      job.trigger(tag = tag)
      latest_run = job.last_n_runs(1)

      if not _wait:
        # even though we are returning the run_id, this cannot be trusted when making parallel call because of
        # the race condition on top run, which is the case when .map() method is called on the operator. 
        # return tag, latest_run["id"]
        return tag, latest_run["id"]

      # if required wait else just return the 
      max_retries = latest_run['resource']['max_retries']

      max_polls = 600  # about 10 mins
      poll_count = 0
      while latest_run["retry_count"] <= max_retries:
        # create a polling loop
        logger.debug(f"[{poll_count:04d}/{max_polls:04d}] [{latest_run['retry_count']}/{max_retries}] {latest_run['status']}")
        if poll_count > max_polls:
          raise TimeoutError(f"Run {latest_run['id']} timed out after {max_polls} polls")
        if latest_run["status"] == "COMPLETED":
          break
        elif latest_run["retry_count"] == max_retries and latest_run["status"] == "ERROR":
          raise RuntimeError(f"Run {latest_run['id']} failed after {max_retries} retries")
        latest_run = job.last_n_runs(1)
        poll_count += 1
        sleep(4)

      if latest_run["status"] == "ERROR":
        raise Exception(f"Run failed after {max_retries} retries")

      # assuming everything went well we should have a file at /{job_id}/return
      obj = relic.get_object(f"{job_id}/return_{tag}")
      return obj
    
    # create the class and override some values to make more sense
    _op = cls()
    _op.__qualname__ = "job_" + job_id
    _op.forward = forward
    _op._op_type = ospec.OperatorType.JOB
    _op._op_spec = ospec._JobSpec(
      job_id = job_id,
      rpc_fn_name = job_id,
      job = job,
      workspace_id = workspace_id,
    )

    return _op

  @classmethod
  def from_serving(cls, url: str, token: str):
    """Latch to an existing serving operator

    Args:
      url (str): The URL of the serving
      token (str): The token to access the deployment, get it from settings.
    """
    logger.debug(f"Latching to serving: {url}")

    # now we can run a NBX-Let either on a Pod or on the Build instance, so we can check the URL
    # once and make a judgement based on that

    session = requests.Session()
    if re.match("https:\/\/api\.nimblebox\.ai\/(\w+)\/", url):
      # this is deployment on a Pod
      session.headers.update({"NBX-KEY": token})
      serving_id = url.split("/")[-1]
    elif re.match("https:\/\/(\w+)-(\w+)\.build([\.rc]+)*\.nimblebox\.ai\/", url):
      # this is deployment on a Build instance, there's a catch though without knowing the
      session.headers.update({
        "NBX-TOKEN": token,
        "X-NBX-USERNAME": secret.get("username"),
      })
      serving_id = url
    else:
      raise ValueError(f"Invalid URL: {url}")

    openapi_url = f"{url}openapi.json" # url has as trailing slash
    r = session.get(openapi_url)
    try:
      data = r.json()
    except:
      logger.error(r.content)
      raise

    REQ = type("REQ", (object,), {})

    # create a function to input spec mapper and so the generic forward method can be 
    fn_spec = {}
    for p, v in data["paths"].items():
      if p.startswith("/method_"):
        fn_name = p[8:]
        v = v["post"]
        # op_id = v["operationId"]
        ref_path = v["requestBody"]["content"]["application/json"]["schema"]["$ref"].split("/")[1:]
        out = data
        for p in ref_path:
          out = out[p]
        args = out["properties"]
        fn_spec[fn_name] = args
      elif p == "/forward":
        fn_name = "forward"
        v = v["post"]
        # op_id = v["operationId"]
        ref_path = v["requestBody"]["content"]["application/json"]["schema"]["$ref"].split("/")[1:]
        out = data
        for p in ref_path:
          out = out[p]
        args = out["properties"]
        fn_spec[fn_name] = args

    serving_stub = SpecSubway.from_openapi(data, _url = url, _session = session)

    # define the forward function for this serving operator, the objective is that this will be able to handle
    # args, kwargs just like how it works on the local machine
    def forward(method, *args, **kwargs):
      if method in fn_spec:
        # check in the kwargs if we have any arguments to pass
        args_dict = {k: v.get("default", REQ) for k, v in fn_spec[method].items()}
        _data = {} # the json that will be sent over
        for i, (k, v) in enumerate(args_dict.items()):
          if len(args) == i:
            break
          _data[k] = args[i]
        for k, v in kwargs.items():
          _data[k] = v

        # client-side check for unknown vars
        unknown_args = set()
        for k, v in _data.items():
          if k not in args_dict:
            unknown_args.add(k)
        if len(unknown_args):
          raise ValueError(f"Unknown arguments: {unknown_args}")

        missing_args = set()
        for k, v in args_dict.items():
          if k not in _data and v == REQ:
            missing_args.add(k)
        if len(missing_args):
          raise ValueError(f"Missing required arguments: {missing_args}")

        # this is a simple method call
        if method == "forward":
          fn = "forward"
        elif method in fn_spec:
          fn = f"method_{method}"

        # serialize everything to b64
        for k, v in _data.items():
          _data[k] = U.py_to_bs64(v)
      else:
        _data = {"rpc_name": method, "key": U.py_to_bs64(args[0])}
        if len(args) > 1:
          _data["value"] = U.py_to_bs64(args[1])
        fn = "nbx_py_rpc"

      # now we can call the function
      data = serving_stub.u(fn)(**_data)

      # convert to usable values
      if not data["success"]:
        raise Exception(data["message"])
      value = U.py_from_bs64(data["value"])
      return value

    # call the stub and get details of the operator
    try:
      data = serving_stub.who_are_you()
    except AttributeError as e:
      logger.error(f"Error: {e}")
      logger.error("Unable to connect to the serving, you are probably not connected to a nbox serving")
      raise ValueError("Unable to connect to the serving, you are probably not connected to a nbox serving")

    # create the class and override some values to make more sense
    _op = cls()
    _op.__qualname__ = "serving_" + data["name"]

    _op.forward = forward
    _op._op_type = ospec.OperatorType.SERVING
    _op._op_spec = ospec._ServingSpec(
      serving_id = serving_id,
      rpc_fn_name = data["name"],
      fn_spec = fn_spec,
      workspace_id = "unknown--",
    )
    return _op

  @classmethod
  def fn(cls):
    """Wraps a function or class as an Operator, so you can use all the same methods as Operator"""
    def wrap(fn):
      if type(fn) == type(wrap): # lol the quick hack
        # this is a wrapped function to be run as a job
        op = cls()
        op = op._fn(fn)
        return op
      elif type(fn) == type(Operator):
        # this is a little bit tricky since the initialisation of the object has be be done
        # by the user later and thus we wrap another function which actually initialises the
        # object
        def cls_init(*args, **kwargs):
          op = cls()
          op = op._cls(fn, *args, **kwargs)
          return op
        return cls_init
    return wrap

  def _cls(self, fn, *args, **kwargs):
    """Do not use directly, use `@operator` decorator instead. Utility to wrap a class as an operator"""
    obj = fn(*args, **kwargs)
    # override the file this is necessary for getting the remote operations running
    self.__file__ = inspect.getfile(fn)
    self.__doc__ = obj.__doc__
    self.__qualname__ = "cls_" + obj.__class__.__qualname__
    self._op_type = ospec.OperatorType.WRAP_CLS
    self._op_spec = ospec._WrapClsSpec(
      cls_name = obj.__class__.__qualname__,
      wrap_obj = obj,
      init_ak = (args, kwargs)
    )

    return self

  @classmethod
  def from_class(cls, obj):
    """Wraps an initialised class as an operator, so you can use all the same methods as Operator"""
    op = cls()
    op.__file__ = inspect.getfile(obj.__class__)
    op.__doc__ = obj.__doc__
    op.__qualname__ = "cls_" + obj.__class__.__qualname__
    op._op_type = ospec.OperatorType.WRAP_CLS
    op._op_spec = ospec._WrapClsSpec(
      cls_name = obj.__class__.__qualname__,
      wrap_obj = obj,
      init_ak = ((), {})
    )
    return op

  def _fn(self, fn):
    """Do not use directly, use `@operator` decorator instead. Utility to wrap a function as an operator"""
    self.forward = fn # override the forward function
    # override the file this is necessary for getting the remote operations running
    self.__file__ = inspect.getfile(fn)
    self.__doc__ = fn.__doc__
    self.__qualname__ = "fn_" + fn.__name__
    self._op_type = ospec.OperatorType.WRAP_FN
    self._op_spec = ospec._WrapFnSpec(fn_name = fn.__name__, wrap_obj = fn)

    return self

  @classmethod
  def from_fn(cls, fn):
    """Wraps a function as an operator, so you can use all the same methods as Operator"""
    op = cls()
    op = op._fn(fn)
    return op
    

  # /mixin

  # python state modification /

  """The functions below are part of a larger effort to make Operator interact better with the underlying
  user classes and functions. Most Importantly we are borrowing the ray concept that functions map to jobs
  and classes map to actors. We are also extending it a little bit to make it more pythonic in ways by adding
  suppoprt for multiple inbuilt python object methods like __getitem__, __setitem__ etc. So in order to make
  it work we are adding those methods below as well and each method will based on its type take a call on
  what to do.
  """

  def __setattr__(self, key, value: 'Operator'):
    if (
      isinstance(value, Operator) and
      hasattr(value, "_op_type") and self._op_type == ospec.OperatorType.UNSET and
      hasattr(self, "_operators")
    ):
      # print("Adding:", self.__class__.__qualname__, key, value.__class__.__qualname__, isinstance(value, Operator), id(value), id(self))
      self._operators[key] = value
    self.__dict__[key] = value

  def __getattr__(self, key):
    # if you try to access any attribute of self outside the conditionals it will cause a recursion error
    if self._op_type == ospec.OperatorType.SERVING:
      if key in self._op_spec.fn_spec:
        return partial(self.forward, key)
      return self.forward("__getattr__", key)
    elif self._op_type == ospec.OperatorType.WRAP_CLS:
      return getattr(self._op_spec.wrap_obj, key)
    elif self._op_type == ospec.OperatorType.UNSET and key == "__qualname__":
      return self.__class__.__qualname__
    raise AttributeError(f"{key}")

  def __getitem__(self, key):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__getitem__", key)
    if self._op_type in [ospec.OperatorType.WRAP_FN, ospec.OperatorType.WRAP_CLS]:
      return self._op_wrap[key]
    raise KeyError(f"{key}")

  def __setitem__(self, key, value):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__setitem__", key, value)
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      self._op_wrap[key] = value
      return
    raise KeyError(f"{key}")

  def __delitem__(self, key):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__delitem__", key)
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      del self._op_wrap[key]; return
    raise KeyError(f"{key}")

  def __iter__(self):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__iter__")
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      return iter(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __next__(self):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__next__")
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      return next(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __len__(self):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__len__")
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      return len(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __contains__(self, key):
    if self._op_type == ospec.OperatorType.SERVING:
      return self.forward("__contains__", key)
    if self._op_type in [ospec.OperatorType.WRAP_CLS]:
      return key in self._op_wrap
    raise ValueError(f"Operator cannot iterate")

  # / python state modification

  def propagate(self, **kwargs):
    """Set kwargs for each child in the Operator"""
    for k, v in kwargs.items():
      setattr(self, k, v)
    for c in self._operators.values():
      c.propagate(**kwargs)

  def thaw(self, job: JobProto):
    """Load JobProto into this Operator"""
    nodes = job.dag.flowchart.nodes
    edges = job.dag.flowchart.edges
    for _id, node in nodes.items():
      name = node.name
      if name.startswith("self."):
        name = name[5:]
      if hasattr(self, name):
        op: 'Operator' = getattr(self, name)
        op.propagate(
          node = node,
          source_edges = list(filter(
            lambda x: edges[x].target == node.id, edges.keys()
          ))
        )

  def _named_operators(self, memo = None, prefix: str = '', remove_duplicate: bool = True):
    r"""Returns an iterator over all modules in the network, yielding both the name of the module
    as well as the module itself."""
    if memo is None:
      memo = set()
    if self not in memo:
      if remove_duplicate:
        memo.add(self)
      yield prefix, self
      for name, module in self._operators.items():
        if module is None:
          continue
        submodule_prefix = prefix + ('.' if prefix else '') + name
        for m in module._named_operators(memo, submodule_prefix, remove_duplicate):
          yield m

  def __call__(self, *args, **kwargs):
    # check if calling is allowed based on the type
    if self._op_type == ospec.OperatorType.SERVING:
      # the fn_spec is set during the .from_serving() classmethod
      if len(self._op_spec.fn_spec) == 1 and "forward" in self._op_spec.fn_spec:
        # this means that an ospec.OperatorType.UNSET or ospec.OperatorType.WRAP_FN are on the other side of the pipe
        return self.forward("forward", *args, **kwargs)
      else:
        # this means that ospec.OperatorType.WRAP_CLS is on the other side of the pipe
        raise ValueError(f"Cannot call servings directly, will interfere with nbox")
    elif self._op_type == ospec.OperatorType.WRAP_CLS:
      raise ValueError(f"Cannot call class wrappers directly, will interfere with nbox")

    input_dict = {}
    logger.debug(f"Calling operator '{self.__class__.__name__}': {self.node.id}")
    _ts = SimplerTimes.get_now_pb()
    self.node.run_status.CopyFrom(RunStatus(start = _ts, inputs = {k: str(type(v)) for k, v in input_dict.items()}))
    if self._tracer != None:
      self._tracer(self.node)
    # ---- USER SEPERATION BOUNDARY ---- #

    with log_latency(f"{self.__class__.__name__}-forward"):
      out = self.forward(*args, **kwargs)

    # ---- USER SEPERATION BOUNDARY ---- #
    outputs = {}
    logger.debug(f"Ending operator '{self.__class__.__name__}': {self.node.id}")
    _ts = SimplerTimes.get_now_pb()
    self.node.run_status.MergeFrom(RunStatus(end = _ts, outputs = {k: str(type(v)) for k, v in outputs.items()}))
    if self._tracer != None:
      self._tracer(self.node)

    # if user has enabled _tracking, then we will store the input, output values as well
    return out

  def forward(self):
    raise NotImplementedError("User must implement forward()")

  # nbx/
  def _get_dag(self) -> DAG:
    """Get the DAG for this Operator including all the nested ones."""
    dag = get_nbx_flow(self.forward)
    all_child_nodes = {}
    all_edges = {}
    for child_id, child_node in dag.flowchart.nodes.items():
      name = child_node.name
      if name.startswith("self."):
        name = name[5:]
      operator_name = "CodeBlock" # default
      cls_item = getattr(self, name, None)
      if cls_item is not None and cls_item.__class__.__base__ == Operator:
        # this node is an operator
        operator_name = cls_item.__class__.__name__
        child_dag: DAG = cls_item._get_dag() # call this function recursively

        # update the child nodes with parent node id
        for _child_id, _child_node in child_dag.flowchart.nodes.items():
          if _child_node.parent_node_id == "":
            _child_node.parent_node_id = child_id # if there is already a parent_node_id set is child's child
          all_child_nodes[_child_id] = _child_node
          all_edges.update(child_dag.flowchart.edges)
      
      # update the child nodes name with the operator name
      child_node.operator = operator_name

    # update the root dag with new children and edges
    _nodes = {k:v for k,v in dag.flowchart.nodes.items()}
    _edges = {k:v for k,v in dag.flowchart.edges.items()}
    _nodes.update(all_child_nodes)
    _edges.update(all_edges)

    # because updating map in protobuf is hard
    dag.flowchart.CopyFrom(Flowchart(nodes = _nodes, edges = _edges))
    return dag

  def deploy(
    self,
    group_id: str,
    deployment_type: str = None,
    resource: Resource = None,
    ignore_patterns: List[str] = [],
    workspace_id: str = "",
    *,
    _unittest = False,
    _include_pattern = [],
  ) -> 'Operator':
    """Uploads relevant files to the cloud and deploys as a batch process or and API endpoint, returns the relevant
    `.from_job()` or `.from_serving` Operator. This uploads the entire folder where the caller file is located.
    In which case having a `.nboxignore` and `requirements.txt` will be honoured.

    Args:
      group_id (str): The Job/Deploy id to deploy to.
      deployment_type (str, optional): The deployment type. Defaults to None.
      resource (Resource, optional): The resource to use for deployment. Defaults to None.
      ignore_patterns (List[str], optional): The patterns to ignore. Defaults to [].
    """

    if not workspace_id:
      workspace_id = secret.get(ConfigString.workspace_id)
      logger.info(f"Using workspace_id: {workspace_id}")

    # go over reasonable checks for deployment
    if deployment_type == None:
      if self._op_type in [ospec.OperatorType.WRAP_CLS, ospec.OperatorType.SERVING]:
        deployment_type = "serving"
      else:
        deployment_type = "job"
    if self._op_type == ospec.OperatorType.WRAP_CLS and deployment_type != "serving":
      raise ValueError(f"Cannot deploy a class as a job, only as a serving")
    if deployment_type not in ospec.OperatorType._valid_deployment_types():
      raise ValueError(f"Invalid deployment type: {deployment_type}. Must be one of {ospec.OperatorType._valid_deployment_types()}")
    if deployment_type == ospec.OperatorType.SERVING:
      if group_id is None:
        raise ValueError("pass deployment_id as group_id for serving deployment")

    # get the filepath and name to import for convience
    fp, folder, file, name = ospec.get_operator_location(self)
    logger.info(f"Deployment Type: {deployment_type}")

    # create a temporary directory to store all the files
    # copy over all the files and wait for it, else the changes below won't be reflected
    # dude damn, how did I not know this: https://dev.to/ackshaey/macos-vs-linux-the-cp-command-will-trip-you-up-2p00
    # so basically till now all we know is that `from fp import name` and `name()`. This process can now
    # be used to create a custom `nbx_user.py` file from which `exe.py` can import the three functions
    # and use them the conventional way. the three functions could simply be a router and the real operator
    # can be imported from the orginal file.
    init = ''
    if self._op_type in [ospec.OperatorType.UNSET, ospec.OperatorType.WRAP_CLS]:
      init = '()'

    with open(U.join(folder, "nbx_user.py"), "w") as f:
      f.write(f'''# Autogenerated for .deploy() call
from nbox.messages import read_file_to_binary
from nbox.hyperloop.jobs.job_pb2 import Resource

from {file.split('.')[0]} import {name}

def get_op(*_, **__):
  # takes no input since programtically generated returns the exact object
  out = {name}{init}
  return out

get_resource = lambda: read_file_to_binary('.nbx_core/resource.pb', message = Resource())
get_schedule = lambda: None
''')

    # create a requirements.txt file if it doesn't exist with the latest nbox version
    if not os.path.exists(U.join(folder, "requirements.txt")):
      with open(U.join(folder, "requirements.txt"), "w") as f:
        f.write(f'nbox[serving]=={__version__}\n')

    # create a .nboxignore file if it doesn't exist, this will tell all the files not to be uploaded to the job pod
    _igp = set(FN_IGNORE + ignore_patterns)
    _igp -= set(_include_pattern)
    if not os.path.exists(U.join(folder, ".nboxignore")):
      with open(U.join(folder, ".nboxignore"), "w") as f:
        f.write("\n".join(_igp))
    else:
      with open(U.join(folder, ".nboxignore"), "r") as f:
        _igp = _igp.union(set(f.read().splitlines()))
      with open(U.join(folder, ".nboxignore"), "w") as f:
        f.write("\n".join(_igp))

    # just create a resource.pb, if it's empty protobuf will work it out
    nbx_folder = U.join(folder, ".nbx_core")
    os.makedirs(nbx_folder, exist_ok = True)
    write_binary_to_file(resource or DEFAULT_RESOURCE, file = U.join(nbx_folder, "resource.pb"))

    if _unittest:
      return

    init_folder = f"{fp.replace('.py', '')}:{name}"
    logger.debug(f"Upload init_folder: {init_folder}")

    if deployment_type == ospec.OperatorType.JOB:
      out = Job.upload(
        init_folder = init_folder,
        id = group_id,
        _ret = True
      )
      return self.from_job(job_id = out.id)
    elif deployment_type == ospec.OperatorType.SERVING:
      out = Serve.upload(
        init_folder = init_folder,
        id = group_id,
        _ret = True
      )

      # get the serving object
      stub = nbox_ws_v1.deployments.u(out.id)

      # we will poll here till the model is not ready and return the latched operator
      done = False
      while not done:
        sleep(1)
        data = stub()
        this_model = list(filter(lambda x: x["id"] == out.model_id, data["models"]))[0]
        status = this_model["status"]
        logger.debug(f"status of model: {status}")
        if status == "deployment.failed":
          raise Exception("Deployment failed")
        done = status == "deployment.ready"

      token = data["deployment"]["api_key"]
      api_url = data["deployment"]["api_url"]
      model_url = api_url + out.model_id + "/"
      logger.info("model_url: " + model_url)
      return self.from_serving(model_url, token = token)

  def map(self, inputs: Union[List[Any], Tuple[Any]], workers: int = -1) -> Iterable[Any]:
    """Take the same logic and apply it to a list of inputs, different from star_map in that it
    takes in different logic and applied different inputs. Returns results in the same order as
    inputs.
    
    In the current version runs as many workers as are the inputs."""
    if self._op_type == ospec.OperatorType.SERVING:
      raise NotImplementedError("What does a map call really mean for an API endpoint? Like parallel calls, try mapping a function.")
    elif self._op_type == ospec.OperatorType.WRAP_CLS:
      raise NotImplementedError("What does a map call really mean for a class?")

    # this another big ass function that manages this call for all the different types of ospec.OperatorTypes
    if len(inputs) > 10:
      raise RuntimeError(f"Too many maps: {len(inputs)}, current limit is 10")
    workers = len(inputs) if workers == -1 else workers
    inputs = [[x] for x in inputs]

    # the main if/else tree
    if self._op_type in [ospec.OperatorType.UNSET, ospec.OperatorType.WRAP_FN]:
      import sys
      from subprocess import Popen
      FOLDER = ".nbx-para-logs"
      os.makedirs(FOLDER, exist_ok = True)

      # here's how this thing does a map on local machines:
      # * we create an exe.py object in the current folder, serialise and store all the inputs in the local relics
      #   start a process manager that waits till all the processes are done and files are ready for consumption
      #   all the logs (stdout, stderr) of the execution will be written in separate files.
      # * the current parent process only cares about ensuring that the process runs correctly, and not about the
      #   logic to be executed by the workers.
      # * instead of writing to the RelicsLocal we directly write to file which helps us avoid lack of locks in relics.
      #   all the files will be written in the .nbx-para-logs folder.

      # create the exe.py file
      fp, folder, file, name = ospec.get_operator_location(self)
      exe_file = U.join(folder, "exe.py")
      init = ''
      if self._op_type == ospec.OperatorType.UNSET:
        init = '()'
      with open(exe_file, "w") as f:
        f.write(f"""
import os
os.environ["PYTHONUNBUFFERED"] = "true" # so print comes when it should come
from nbox.lib.dist import LocalNBXLet
from {file.split('.')[0]} import {name}

if __name__ == "__main__":
  tag = os.getenv("NBOX_TAG_ID", None)
  if tag is None:
    raise ValueError("NBOX_TAG_ID not set")
  op = LocalNBXLet({name}{init}, os.path.join("{FOLDER}",tag+"_input"), os.path.join("{FOLDER}",tag+"_output"))
  print(op)
  op()
""".strip())

      run_tags = [U.get_random_name(True).split("-")[0] for _ in range(len(inputs))]
      U.threaded_map(
        lambda tag, inp: U.to_pickle(inp, U.join(FOLDER, tag + "_input")),
        [[x,inp] for x,inp in zip(run_tags, inputs)],
        _name = "write_inputs",
      )

      processes = []
      run_tag_to_input_idx = {}
      proc_ids = []
      open_files = []
      for idx, tag in enumerate(run_tags):
        run_tag_to_input_idx[tag] = idx
        stdout = open(f".nbx-para-logs/{tag}.log", "w")
        proc = Popen(
          args = [sys.executable, exe_file],
          bufsize = 2 << 9, # 1024 bytes buffer when writing the files
          env = {"NBOX_TAG_ID": tag},
          stdout = stdout,
          stderr = stdout,
          shell = False
        )
        processes.append(proc)
        proc_ids.append(proc.pid)
        open_files.append(stdout)

      # wait till all the processes are done
      while True:
        sleep(1)
        status = [p.poll() for p in processes]
        success = [x == 0 for x in status]
        errored = [x is not None and x != 0 for x in status]
        completed = sum([s is not None for s in status])
        logger.debug(f"total: {len(status)}, success: {sum(success)}, errored: {sum(errored)}, completed: {completed}")
        if completed == len(inputs):
          break
      
      # get the results
      results = U.threaded_map(lambda x: U.from_pickle(U.join(FOLDER, x + "_output")), [[x] for x in run_tags], _name = "read_outputs")

      # clean up
      for f in open_files:
        f.close()
      return results

    elif self._op_type == ospec.OperatorType.JOB:
      # now this is simple enough, we need to implement a waiting queue that's it
      _start_time = monotonic()
      run_tags = []
      run_tag_to_input_idx = {}

      # though the threadpool is a faster method, cannot trust the run_id because of the race condition on top run
      with ThreadPoolExecutor(min(len(inputs), 10)) as exe:
        trigger = lambda i, x: [i, self(*x, _wait = False)]
        results = {exe.submit(trigger, i, x) for i,x in enumerate(inputs)}
        for future in as_completed(results):
          try:
            i, (tag, run_id) = future.result()
            run_tags.append(tag)
            run_tag_to_input_idx[tag] = i
          except Exception as e:
            raise Exception(e)

      run_tag_to_input_idx = {k:v for k,v in sorted(run_tag_to_input_idx.items(), key = lambda x: x[1])}
      sleep(1) # sleep for a bit before requesting the webserver, this sleep matters now that we have the threadpool
      runs = self._op_spec.job.last_n_runs(len(inputs))
      run_ids = [x["id"] for x in runs]

      # print("run_tags:", run_tags)
      # print("run_ids:", run_ids)
      # print("run_tag_to_input_idx:", run_tag_to_input_idx)
      # print("inputs:", inputs, len(inputs))

      proc_hash = U.hash_(",".join(run_tags), "sha256")[:8]
      logger.info(f"Waiting for {len(run_ids)} processes to finish, this may take a while...")
      html_path = U.join(U.env.NBOX_HOME_DIR(), ".cache", f"{proc_hash}.html")

      def _write_html_page(data, headers = ["created_at", "end_time", "input_id", "run_id", "status", "tag", "files"]):
        with open(U.join(U.folder(__file__), "assets", "run_status.jinja"), "r") as src, open(html_path, "w") as dst:
          import jinja2
          template = jinja2.Template(src.read())
          dst.write(template.render(data = data, headers = headers))

      _write_html_page([[] for _ in range(len(inputs))])

      run_status = {rid: None for rid in run_ids}
      active_runs = {rid for rid, done in run_status.items() if not done}
      pbar = tqdm(total = len(inputs), desc = f"Waiting for runs ({html_path})")
      relic = Relics("cache", workspace_id = self._op_spec.workspace_id)

      def _update_runs():
        # get all runs, filter those in this invocation, update the status
        files_ = {}
        for t in run_tags:
          files_[t] = (
            relic.has(f"{self._op_spec.job_id}/return_{tag}"),
          )
        out = self._op_spec.job.last_n_runs(len(inputs))
        # print(out)
        if not isinstance(out, (list, tuple)):
          out = [out]
        proc_runs = list(filter(lambda x: x["id"] in run_ids, out))
        data = [
          [
            x["created_at"],
            x["end_time"],
            run_tag_to_input_idx[t],
            x["id"],
            x["status"],
            t,
            sum(files_[t])
          ] for x, t in zip(proc_runs, run_tags)
        ]
        data = sorted(data, key = lambda x: x[2])
        _write_html_page(data)

        _active_runs = {x["id"] for x in proc_runs if x["status"] not in ["COMPLETED", "ERROR"]}
        pbar.update(len(active_runs) - len(_active_runs))

        all_files_ready = sum([sum(files_[t]) for t in run_tags]) == len(run_tags)

        return _active_runs, all_files_ready

      while len(active_runs) > 0:
        active_runs, all_files_ready = _update_runs()
        if all_files_ready:
          break
        sleep(1)

      pbar.close()

      # load the results in memory in the exact order of the inputs
      results = []
      for tag, _idx in run_tag_to_input_idx.items():
        obj = relic.get_object(f"{self._op_spec.job_id}/return_{tag}")
        results.append(obj)

      _end_time = monotonic()

      logger.info(f"Finished waiting for {len(run_ids)} processes to finish in {_end_time - _start_time} seconds")
      return results

  # /nbx

operator = Operator.fn # convinience



from enum import Enum
from typing import Union
from .jobs.nbox_ws_pb2 import JobRequest, ListJobsRequest, UpdateJobRequest
from .deploy.serve_pb2 import ServingListRequest, ServingRequest, UpdateServingRequest, ModelListRequest, ModelRequest

# To be used as TYPE.ACTION_[RESOURCE]_STATUS
class Actions(Enum):
    # Common actions
    INIT = "INIT_DONE"
    # CODE_UPLOADED = "UPLOAD_CODE_DONE"
    UPDATED_STATUS = "UPDATE_STATUS_DONE"
    UPDATE_RESOURCES_INIT = "UPDATE_RESOURCES_INIT"
    UPDATE_RESOURCES_DONE = "UPDATE_RESOURCES_DONE"
    UPDATE_RESOURCES_FAIL = "UPDATE_RESOURCES_FAIL"
    UPDATE_INFO_INIT = "UPDATE_INFO_INIT"
    UPDATE_INFO_DONE = "UPDATE_INFO_DONE"
    UPDATE_INFO_FAIL = "UPDATE_INFO_FAIL"
    SOURCE_CODE_UPDATE_DONE = "UPDATE_SOURCE_CODE_DONE"
    SOURCE_CODE_UPDATE_FAIL = "UPDATE_SOURCE_CODE_FAIL"
    DELETE_INIT = "DELETE_INIT"
    DELETE_DONE = "DELETE_DONE"
    DELETE_FAIL = "DELETE_FAIL"
    UPDATE_FEATURE_GATES_INIT = "UPDATE_FEATURE_GATES_INIT"
    UPDATE_FEATURE_GATES_DONE = "UPDATE_FEATURE_GATES_DONE"
    UPDATE_FEATURE_GATES_FAIL = "UPDATE_FEATURE_GATES_FAIL"
    UPDATE_AUTH_INFO_INIT = "UPDATE_AUTH_INFO_INIT"
    UPDATE_AUTH_INFO_DONE = "UPDATE_AUTH_INFO_DONE"
    UPDATE_AUTH_INFO_FAIL = "UPDATE_AUTH_INFO_FAIL"
    K8S_CREATE_INIT = "K8S_CREATE_INIT"
    K8S_CREATE_DONE = "K8S_CREATE_DONE"
    K8S_CREATE_FAIL = "K8S_CREATE_FAIL"

    PAUSE_INIT = "PAUSE_INIT"
    PAUSE_DONE = "PAUSE_DONE"
    PAUSE_FAIL = "PAUSE_FAIL"
    RESUME_INIT = "RESUME_INIT"
    RESUME_DONE = "RESUME_DONE"
    RESUME_FAIL = "RESUME_FAIL"
    RESCHEDULE_INIT = "RESCHEDULE_INIT"
    RESCHEDULE_DONE = "RESCHEDULE_DONE"
    RESCHEDULE_FAIL = "RESCHEDULE_FAIL"
    TRIGGER_INIT = "TRIGGER_INIT"
    TRIGGER_DONE = "TRIGGER_DONE"
    TRIGGER_FAIL = "TRIGGER_FAIL"
    PREPARE_RUN_INIT = "PREPARE_RUN_INIT"
    PREPARE_RUN_DONE = "PREPARE_RUN_DONE"
    PREPARE_RUN_FAIL = "PREPARE_RUN_FAIL"

    # Serving Actions
    UNSPECIFIED = "UNSPECIFIED"
    PIN_INIT = "PIN_INIT"
    PIN_DONE = "PIN_DONE"
    PIN_FAIL = "PIN_FAIL"
    UNPIN_INIT = "UNPIN_INIT"
    UNPIN_DONE = "UNPIN_DONE"
    UNPIN_FAIL = "UNPIN_FAIL"
    UPDATE_SCALE_INIT = "UPDATE_SCALE_INIT"
    UPDATE_SCALE_DONE = "UPDATE_SCALE_DONE"
    UPDATE_SCALE_FAIL = "UPDATE_SCALE_FAIL"
    UPDATE_ENDPOINT_ACCESS_KEY_INIT = "UPDATE_ENDPOINT_ACCESS_KEY_INIT"
    UPDATE_ENDPOINT_ACCESS_KEY_DONE = "UPDATE_ENDPOINT_ACCESS_KEY_DONE"
    UPDATE_ENDPOINT_ACCESS_KEY_FAIL = "UPDATE_ENDPOINT_ACCESS_KEY_FAIL"

    @classmethod
    def from_string(cls, value: str) -> "Actions":
        try:
            return cls(value)
        except ValueError:
            return cls.UNSPECIFIED


class Type(Enum):
    JOB = "JOB"
    SERVING = "SERVING"
    MODEL = "MODEL"


JobsVersionType = dict[str, Union[str, list[str]]]


OneClickRequestTypes = Union[
    JobRequest,
    ListJobsRequest,
    UpdateJobRequest,
    ServingListRequest,
    ServingRequest,
    UpdateServingRequest,
    ModelRequest,
    ModelListRequest,
]



import abc
from typing import Any, Callable, Iterator
import grpc
from ..constants import OneClickRequestTypes


class ClientInterceptorReturnType(grpc.Call, grpc.Future):
    """Return type for the ClientInterceptor.intercept method."""

    pass


class ClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
    metaclass=abc.ABCMeta,
):
    """Base class for client-side interceptors.
    To implement an interceptor, subclass this class and override the intercept method.
    """

    @abc.abstractmethod
    def intercept(
        self,
        method: Callable,
        request_or_iterator: Any,
        call_details: grpc.ClientCallDetails,
    ) -> ClientInterceptorReturnType:
        """Override this method to implement a custom interceptor.
        This method is called for all unary and streaming RPCs. The interceptor
        implementation should call `method` using a `grpc.ClientCallDetails` and the
        `request_or_iterator` object as parameters. The `request_or_iterator`
        parameter may be type checked to determine if this is a singluar request
        for unary RPCs or an iterator for client-streaming or client-server streaming
        RPCs.
        Args:
            method: A function that proceeds with the invocation by executing the next
                interceptor in the chain or invoking the actual RPC on the underlying
                channel.
            request_or_iterator: RPC request message or iterator of request messages
                for streaming requests.
            call_details: Describes an RPC to be invoked.
        Returns:
            The type of the return should match the type of the return value received
            by calling `method`. This is an object that is both a
            `Call <https://grpc.github.io/grpc/python/grpc.html#grpc.Call>`_ for the
            RPC and a
            `Future <https://grpc.github.io/grpc/python/grpc.html#grpc.Future>`_.
            The actual result from the RPC can be got by calling `.result()` on the
            value returned from `method`.
        """
        return method(request_or_iterator, call_details)  # pragma: no cover

    def intercept_unary_unary(
        self,
        continuation: Callable,
        call_details: grpc.ClientCallDetails,
        request: Any,
    ):
        """Implementation of grpc.UnaryUnaryClientInterceptor.
        This is not part of the grpc_interceptor.ClientInterceptor API, but must have
        a public name. Do not override it, unless you know what you're doing.
        """
        return self.intercept(_swap_args(continuation), request, call_details)

    def intercept_unary_stream(
        self,
        continuation: Callable,
        call_details: grpc.ClientCallDetails,
        request: Any,
    ):
        """Implementation of grpc.UnaryStreamClientInterceptor.
        This is not part of the grpc_interceptor.ClientInterceptor API, but must have
        a public name. Do not override it, unless you know what you're doing.
        """
        return self.intercept(_swap_args(continuation), request, call_details)

    def intercept_stream_unary(
        self,
        continuation: Callable,
        call_details: grpc.ClientCallDetails,
        request_iterator: Iterator[Any],
    ):
        """Implementation of grpc.StreamUnaryClientInterceptor.
        This is not part of the grpc_interceptor.ClientInterceptor API, but must have
        a public name. Do not override it, unless you know what you're doing.
        """
        return self.intercept(_swap_args(continuation), request_iterator, call_details)

    def intercept_stream_stream(
        self,
        continuation: Callable,
        call_details: grpc.ClientCallDetails,
        request_iterator: Iterator[Any],
    ):
        """Implementation of grpc.StreamStreamClientInterceptor.
        This is not part of the grpc_interceptor.ClientInterceptor API, but must have
        a public name. Do not override it, unless you know what you're doing.
        """
        return self.intercept(_swap_args(continuation), request_iterator, call_details)


def _swap_args(fn: Callable[[Any, Any], Any]) -> Callable[[Any, Any], Any]:
    def new_fn(x, y):
        return fn(y, x)

    return new_fn


class IntegrationInjectionInterceptor(ClientInterceptor):
    """An interceptor that injects the integration name and version into the metadata.
    This interceptor is used by the client to inject the integration name and version
    into the metadata. This is used by the server to determine which integration
    version is being used by the client.
    """

    def __init__(self, integration_name: str, integration_id: str, infra_id: str):
        self.integration_name = integration_name
        self.integration_id = integration_id
        self.infra_id = infra_id

    def intercept(
        self,
        method: Callable,
        request: OneClickRequestTypes,
        call_details: grpc.ClientCallDetails,
    ) -> ClientInterceptorReturnType:
        request.integration_data.connection_name = self.integration_name
        request.integration_data.connection_id = self.integration_id
        request.integration_data.infra_id = self.infra_id
        return method(request, call_details)


class ProjectIdInterceptor(ClientInterceptor):
    """An interceptor that injects the project id into the metadata.
    This interceptor is used by the client to inject the project id into the metadata.
    This is used by the server to determine which project id is being used by the
    client.
    """

    def __init__(self, project_id: str, workspace_id: str):
        self.project_id = project_id
        self.workspace_id = workspace_id

    def intercept(
        self,
        method: Callable,
        request: OneClickRequestTypes,
        call_details: grpc.ClientCallDetails,
    ) -> ClientInterceptorReturnType:
        request.auth_info.project_id = self.project_id
        request.auth_info.workspace_id = self.workspace_id
        return method(request, call_details)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yql/assets/rest.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15yql/assets/rest.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb5\x01\n\x04\x45\x63ho\x12/\n\x0bserver_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x11\n\x07message\x18\x02 \x01(\tH\x00\x12\x0f\n\x05\x65rror\x18\x03 \x01(\tH\x00\x12\x13\n\ttraceback\x18\x04 \x01(\tH\x00\x12\x10\n\x08rpc_name\x18\x05 \x01(\t\x12\x12\n\nproto_data\x18\x06 \x01(\t\x12\x15\n\rbase64_string\x18\x07 \x01(\tB\x06\n\x04\x64\x61ta2\\\n\x07yqlCore\x12\x16\n\x06Status\x12\x05.Echo\x1a\x05.Echo\x12 \n\x10ProtoDefinitions\x12\x05.Echo\x1a\x05.Echo\x12\x17\n\x07Predict\x12\x05.Echo\x1a\x05.Echob\x06proto3')



_ECHO = DESCRIPTOR.message_types_by_name['Echo']
Echo = _reflection.GeneratedProtocolMessageType('Echo', (_message.Message,), {
  'DESCRIPTOR' : _ECHO,
  '__module__' : 'yql.assets.rest_pb2'
  # @@protoc_insertion_point(class_scope:Echo)
  })
_sym_db.RegisterMessage(Echo)

_YQLCORE = DESCRIPTOR.services_by_name['yqlCore']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ECHO._serialized_start=59
  _ECHO._serialized_end=240
  _YQLCORE._serialized_start=242
  _YQLCORE._serialized_end=334
# @@protoc_insertion_point(module_scope)



"""
With NimbleBox you can run cluster wide workloads from anywhere. This requires capabilities around distributed computing,
process managements. The code here is tested along with `nbox.Relic` to perform distributed local and cloud processing.

{% CallOut variant="success" label="If you find yourself using this reach out to NimbleBox support." /%}
"""

import os
import json
from pprint import pformat

import nbox.utils as U
from nbox.utils import logger
from nbox.relics import Relics
from nbox.operator import Operator
from nbox.auth import secret, ConfigString
from nbox.nbxlib.tracer import Tracer
from nbox.hyperloop.jobs.job_pb2 import Job
from nbox.nbxlib.serving import serve_operator, SupportedServingTypes as SST
from nbox.messages import read_file_to_binary
from nbox.hyperloop.deploy.serve_pb2 import Model as ModelProto

from nbox.lmao import LMAO_JOB_TYPE_PREFIX, _lmaoConfig, get_lmao_stub, ExperimentConfig
from nbox.sublime.proto.lmao_pb2 import Run, ListProjectsRequest, ListProjectsResponse

# Manager
class LocalNBXLet(Operator):
  def __init__(self, op: Operator, in_key: str, out_key: str):
    super().__init__()
    self.op = op
    self.in_key = in_key
    self.out_key = out_key

  def __repr__(self):
    return f"LocalNBXLet({self.op.__qualname__}, {self.in_key}, {self.out_key})"

  def forward(self):
    x = U.from_pickle(self.in_key)
    y = self.op(*x)
    U.to_pickle(y, self.out_key)


class NBXLet(Operator):
  def __init__(self, op: Operator):
    """The Operator that runs the things on any pod on the NimbleBox Jobs + Deploy platform.
    Name a parody of kubelet, dockerlet, raylet, etc"""
    super().__init__()
    self.op = op

  def run(self):
    """Run this as a batch process"""
    status = Job.Status.ERROR
    try:
      tracer = Tracer()
      if hasattr(self.op._tracer, "job_proto"):
        self.op.thaw(self.op._tracer.job_proto)
      workspace_id = tracer.workspace_id
      logger.info(f"Workspace Id: {workspace_id}")

      # this is important since nbox uses ConfigString.workspace_id place to get workspace_id from while the init_container
      # might place it at a different place. as of this writing, init_container -> "workspace_id" and nbox -> "config.global.workspace_id"
      secret.put(ConfigString.workspace_id, workspace_id, True) 

      job_id = tracer.job_id
      self.op.propagate(_tracer = tracer)

      # get the user defined tag 
      run_tag = os.getenv("NBOX_RUN_METADATA", "")
      logger.info(f"Tag: {run_tag}")

      # in the NimbleBox system we provide tags for each key which essentially tells what is the behaviour
      # of the job. For example if it contains the string LMAO which means we need to initialise a couple
      # of things, or this can be any other job type
      if run_tag.startswith(LMAO_JOB_TYPE_PREFIX):
        # originally we had a strategy to use Relics to store the information about the initialisation and passed args
        # however we are not removing that because we don't want to spend access money when we are anyways storing all
        # the information in the LMAO DB. so now we get the details of the run and get all the information from there.
        _lmao_stub = get_lmao_stub()
        project_id, exp_id = run_tag[len(LMAO_JOB_TYPE_PREFIX):].split("/")
        logger.info(f"Project name (Experiment ID): {project_id} ({exp_id})")

        # get details for the project and the run
        lmao_project = _lmao_stub.list_projects(ListProjectsRequest(
          workspace_id = workspace_id,
          project_id_or_name = project_id
        )).projects[0]
        lmao_run = _lmao_stub.get_run_details(Run(
          workspace_id = workspace_id,
          project_id = project_id,
          experiment_id = exp_id
        ))

        # create the experiment config and set all the values in the _lmaoConfig object that will be passed
        # to the LMAO class despite multiple initialisations, this ensures that when the user also uses LMAO
        # class all the values are already filled up.
        exp_config = ExperimentConfig.from_json(lmao_run.config)
        _lmaoConfig.set(
          project_name = lmao_project.project_name,
          project_id = lmao_run.project_id,
          experiment_id = exp_id,
          save_to_relic = exp_config.save_to_relic,
          enable_system_monitoring = exp_config.enable_system_monitoring,
          store_git_details = True,
        )
        logger.info("LMAO Config:\n" + pformat({
          "kv": _lmaoConfig.kv,
          "experiment": exp_config.to_dict()
        }, compact=True))
        args = ()
        kwargs = exp_config.run_kwargs

      else:
        # check if there is a specific relic for this job
        relic = Relics("cache", workspace_id)
        _in = f"{job_id}/args_kwargs"
        if run_tag:
          _in += f"_{run_tag}"
        logger.info(f"Looking for init.pkl at {_in}")
        if relic.relic is not None and relic.has(_in):
          (args, kwargs) = relic.get_object(_in)
        else:
          args, kwargs = (), {}

      # call the damn thing
      out = self.op(*args, **kwargs)

      # save the output to the relevant place, LMAO jobs are not saved to the relic
      if run_tag.startswith(LMAO_JOB_TYPE_PREFIX):
        logger.info("NBX-LMAO runs does not store function returns in Relics")
      else:
        _out = f"{job_id}/return"
        if run_tag:
          _out += f"_{run_tag}"
        logger.info(f"Saving output to {_out}")
        if relic.relic is not None:
          relic.put_object(_out, out)

      # last step mark as completed
      status = Job.Status.COMPLETED
    except Exception as e:
      U.log_traceback()
    finally:
      logger.info(f"Job {job_id} completed with status {status}")
      if hasattr(tracer, "job_proto"):
        tracer.job_proto.status = status
        tracer._rpc(f"RPC error in ending job {job_id}")
      U.hard_exit_program()

  def serve(self, host: str = "0.0.0.0", port: int = 8000, *, model_name: str = None):
    """Run a serving API endpoint"""

    # Unlike a run above where it is only going to be triggered once and all the metadata is already indexed
    # in the DB, this is not the case with deploy. But with deploy we can get away with something much simpler
    # which is using a more complicated ModelProto.metadata object
    # init_folder = U.env.NBOX_JOB_FOLDER("")
    # if not init_folder:
    #   raise RuntimeError("NBOX_JOB_FOLDER not set")
    # if not os.path.exists(init_folder):
    #   raise RuntimeError(f"NBOX_JOB_FOLDER {init_folder} does not exist")

    # fp_bin = U.join(init_folder, "model_proto.msg")
    # serving_type = SST.NBOX
    # if fp_bin:
    #   model_proto: ModelProto = read_file_to_binary(fp_bin, ModelProto())
    #   logger.info(model_proto)
    #   serving_type = model_proto.metadata.get("serving_type", SST.NBOX)

    try:
      serve_operator(
        op_or_app = self.op,
        # serving_type = serving_type,
        host = host,
        port = port,
        model_name = model_name
      )
    except Exception as e:
      U.log_traceback()
      logger.error(f"Failed to serve operator: {e}")
      U.hard_exit_program()
